#! /usr/exp/bin/perl -w

# mudmap 1.1
#
# Copyright (C) 1996 Raymond Penners <raymondp@stack.urc.tue.nl>
#
# Time-stamp: <96/10/04 11:45:58 raymondp>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# Setup
$prompt_rx = "^>\\s*";
$cmd_quit = "^quit\$";
$cmd_ignore = "^((\\b(l|look|say|read|search|intro|w|who|commune)\\b)|')";
$my_prompt_rx = "^#\\s*";
$my_cmd_comment = "^comment\\s+(\\S)\\s+(.+)\$";
$move_failure_rx = "(What \?|no obvious exit|too (tired|exhausted)|cannot go that way|unfortunately you fail|Sorry, you|The road.*under construction|You are not allowed)";
$exits_rx = "^There .+ obvious exits: ";
$exits_ignore_rx = "^(in|out|up|down|portal|stairs)\$";
$max_coor = 1024;

# Encode directions
%directions = ('e' => 0,
	       'ne' => 1,
	       'n' => 2,
	       'nw' => 3,
	       'w' => 4,
	       'sw' => 5,
	       's' => 6,
	       'se' => 7,

	       'east' => 0,
	       'northeast' => 1,
	       'north' => 2,
	       'northwest' => 3,
	       'west' => 4,
	       'southwest' => 5,
	       'south' => 6,
	       'southeast' => 7
	       );

# Encode movement delta's
@move_delta = ([1, 0],
	       [1, 1],
	       [0, 1],
	       [-1, 1],
	       [-1, 0],
	       [-1, -1],
	       [0, -1],
	       [1, -1]);


# Turns location L into a scalar key
sub location_to_key #@l
{
  my @l = @_;

  return (($l[0] + $max_coor)<<16) | ($l[1] + $max_coor);
}

# Turns KEY into a location
sub key_to_location #$key
{
  my $key = shift;
  return ((($key >> 16) - $max_coor), ($key & 0xffff) - $max_coor);
}

# Print error message
sub eprint #...
{
  my @args = @_;

  print STDERR $0, ": $.: ", @args, "\n";
}

# Converts direction DIR to human-readable string
sub dir_to_str #dir
{
  my ($dir) = @_;
  my ($d, $s);

  foreach $s (keys %directions)
    {
      if ($directions{$s} == $dir)
	{
	  return $s;
	}
    }
  die "can't decode direction";
}


# Converts string S to directions bitfield
sub str_to_dirs #s
{
  my ($s) = @_;

  # Remove "and"
  $s =~ s/\band\b//ig;
  # Erase "," and "."
  $s =~ tr/,\./ /;

  # Split
  my @fields = split /\s+/, $s;
  my  $dirs = 0;
  foreach $d (@fields)
    {
      $d = lc $d;
      if (exists $directions{$d})
	{
	  $dirs |= (1 << $directions{$d});
	}
      elsif (! ($d =~ /$exits_ignore_rx/))
	{
	  &eprint ("ignored unknown direction: \"$d\"");
	}
    }
  
  return $dirs;
}

# Converts directions DIR to strings 
sub dirs_to_str #dir
{
  my ($dir) = @_;
  my $i;
  my $s = "";

  for ($i = 0; $dir!=0; $i++, $dir >>= 1)
    {
      if ($dir & 1)
	{
	  my $t = &dir_to_str ($i);
	  if ($s eq "")
	    {
	      $s = $t;
	    }
	  else
	    {
	      $s = $s . ", " . $t;
	    }
	}
    }
  return $s;
}
  
# Returns the room given location L 
sub room_get #l
{
  my @l = @_;

  ($max_loc[0] < $l[0]) && ($max_loc[0] = $l[0]);
  ($max_loc[1] < $l[1]) && ($max_loc[1] = $l[1]);
  ($min_loc[0] > $l[0]) && ($min_loc[0] = $l[0]);
  ($min_loc[1] > $l[1]) && ($min_loc[1] = $l[1]);
  
  my $key = location_to_key (@l);
  my $r;
  
  if (! (exists $map{$key}))
    {
      $map{$key} = { 'exits' => 0 };
      $r = $map{$key};
    }
  else
    {
      $r = $map{$key};
    }
  if ($location[0]==$l[0] && $location[1]==$l[1])
    {
      $$r{'visited'} = 1;
    }
  return $r;
}

# Sets COMMENT for this room with one-letter identifier ID
sub room_comment #id, #comment
{
  my ($id, $comment) = @_;

  # Add comment
  if (exists $comments{$id})
    {
      &eprint ("comment '$id' redefined",
	       " from \"$comments{$id}\" to \"$comment\"");
    }
  $comments{$id} = $comment;

  # Update room
  my $r = room_get (@location);
  $$r{'comment_id'} = $id;
}

# Add move from room at LOCATION in direction DIR
sub room_add_dir #$dir
{
  my ($dir) = @_;

  # From room
  my $r = room_get (@location);
  $$r{'exits'} |= (1 << $dir);

  # To room
  my @l;
  my $dlt = $move_delta[$dir];
  $l[0] = $location[0] + $$dlt[0];
  $l[1] = $location[1] + $$dlt[1];
  $r = room_get (@l);
  $$r{'exits'} |= (1 << (($dir+4)&7));
}

# Add exits to room
sub room_add_dirs #dirs
{
  my ($dirs) = @_;
  my $i;
  
  for ($i = 0; $dirs!=0; $i++, $dirs >>= 1)
    {
      if ($dirs & 1)
	{
	  &room_add_dir ($i);
	}
    }
}

# Prints map
sub map_print_debug
{
  my $i;
  
  foreach $i (keys %map)
    {
      my @loc = &key_to_location ($i);
      my $exits = $map{$i}{'exits'};
      my $id = $map{$i}{'comment_id'};

      print "$loc[0] x $loc[1]: ";
      if (exists $map{$i}{'comment_id'})
	{
	  print "id=$map{$i}{'comment_id'} ";
	}
      print "exists: ", &dirs_to_str ($exits);
      print "\n";
    }

  foreach $i (keys %comments)
    {
      print "$i -> $comments{$i}\n";
    }
}

# Prints map: ASCII output
sub map_print_ascii
{
  # Top to bottom
  my $y;
  for ($y=$max_loc[1]; $y>=$min_loc[1]; $y--)
    {
      # Two passes per row..
      my $pass;
      for ($pass=0; $pass<2; $pass++)
	{
	  my $x;

	  # Left to right
	  for ($x=$min_loc[0]; $x<=$max_loc[0]; $x++)
	    {
	      my $r = &room_get (($x, $y));
	      my $ex = $$r{'exits'};

	      if ($pass == 0)
		{
		  if ($ex == 0)
		    {
		      print " ";
		    }
		  else
		    {
		      if ($x == $location[0] && $y==$location[1])
			{
			  print "*";
			}
		      else
			{
			  if (exists $$r{'comment_id'})
			    {
			      print $$r{'comment_id'};
			    }
			  elsif (exists $$r{'visited'})
			    {
			      print "O";
			    }
			  else
			    {
			      print "o";
			    }
			}
		    }
		  ($ex & 1) ? print "-" : print " ";
		}
	      else 
		{
		  my $re_ex;
		  
		  # Check room to the east of this one, if possible
		  if ($x+1 <= $max_loc[0])
		    {
		      my $re = &room_get (($x+1, $y));
		      $re_ex = $$re{'exits'};
		    }
		  else
		    {
		      $re_ex = 0;
		    }

		  # South exit 
		  ($ex & (1<<6)) ? print "|" : print " ";

		  # Southeast exit?
		  if ($ex & (1<<7))
		    {
		      # Does room east have southwest exit?
		      ($re_ex & (1<<5)) ? print "X" : print "\\";
		    }
		  else
		    {
		      # Does room east have southwest exit?
		      ($re_ex & (1<<5)) ? print "/" : print " ";
		    }
		}
	    }
	  print "\n";
	}
    }

  # Room comments
  foreach $i (keys %comments)
    {
      print "$i = $comments{$i}\n";
    }
}

# Prints map: mudtrack(TM) compatible output
sub map_print_mudtrack
{
  my $i;
  
  foreach $i (keys %map)
    {
      my @loc = &key_to_location ($i);
      $loc[0] += 50;
      $loc[1] = 50 - $loc[1];
      
      print "$loc[0], $loc[1] ";
      
      my $exits = $map{$i}{'exits'};
      my $id = $map{$i}{'comment_id'};

      my @mt_exits = (2, 4, 0, 6, 3, 1, 5, 7);

      foreach $i (@mt_exits)
	{
	  if ($exits & (1 << $i))
	    {
	      print "1";
	    }
	  else
	    {
	      print "0";
	    }
	  print ",";
	}
      print "0,0 ";

      if (exists $map{$i}{'comment_id'})
	{
	  print ord ($map{$i}{'comment_id'});
	}
      else
	{
	  print 0;
	}
		 
      print "\n";
    }

  # End of mep
  print "-1,0 0,0,0,0,0,0,0,0,0,0 0\n";

  # Comment
  foreach $i (keys %comments)
    {
      print ord($i),",$comments{$i}\n";
    }
}

# Parse LINE
sub parse_my_command #line
{
  my $line = shift @_;

  if ($line =~ $my_cmd_comment)
    {
      &room_comment ($1, $2);
    }
  else
    {
      &eprint ("unknown local command: \"",$line,"\"");
    }
}

# Parse LINE
sub parse_command #line
{
  my $line = shift @_;

  # Got a prompt.. did we move since last prompt?
  if ($moved != -1)
    {
      # Update the old room...
      &room_add_dir ($moved);

      # New room
      @location = @new_location;

      # Set exits if any
      if ($exits != -1)
	{
	  my $r = room_get (@location);
	  # If there are exits which are not in the "obvious exits"
	  # list there is something wrong..
	  if ($$r{'exits'} & (~$exits))
	    {
	      &eprint ("anomaly in space detected!");
	    }
	  &room_add_dirs ($exits);
	  $exits = -1;
	}

      # Done
      $moved = -1;
    }
      
  # My commands
  if ($line =~ $my_prompt_rx)
    {
      &parse_my_command ($');
    }
  # Movements
  elsif (exists $directions{$line})
    {
      $moved = $directions{$line};
      $exits = -1;
      my $dlt = $move_delta[$moved];
      $new_location[0] = $location[0] + $$dlt[0];
      $new_location[1] = $location[1] + $$dlt[1];
    }
  # Quit
  elsif ($line =~ $cmd_quit)
    {
      return -1;
    }
  # Other
  elsif ((! ($line =~ $cmd_ignore)) && ($line ne ""))

    {
      &eprint ("unknown command: \"$line\"");
    }
  return 0;
}
  
# Main
{
  # Parse command line
  use Getopt::Long;
  
  %options = ("help" => 0,
	      "format" => "ascii");
  &GetOptions(\%options, "help","format=s");
  if ($options{"help"})
    {
      print "Usage: $0 [--help] [--format=(mudtrack|ascii)]\n";
      exit 0;
    }

  # Let's start...
  @location = (0, 0);
  @new_location = (0, 0);
  @max_loc = (0, 0);
  @min_loc = (0, 0);
  %map = ();
  %comments = ();
  $moved = -1;
  $exits = -1;

  while (<STDIN>)
    {
      chomp;
      if (/$prompt_rx/)
	{
	  &parse_command ($') && last;
	}
      elsif (/$move_failure_rx/)
	{
	  $moved = -1;
	}
      elsif (/$exits_rx/)
	{
	  $exits = &str_to_dirs ($');
	}
    }
  eval "&map_print_$options{'format'}";
}


